# Generated by Django 6.0.4 on 2026-05-02 12:08

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('test_app', '0002_alter_testmodel_related_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='TestModelCustomObjectId',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('slug', models.CharField(max_length=191, unique=True)),
                ('name', models.CharField(default='v1', max_length=191)),
            ],
        ),
    ]
