# Generated by Django 5.0.1 on 2024-01-30 19:07

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('test_app', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='testmodel',
            name='related',
            field=models.ManyToManyField(blank=True, related_name='+', to='test_app.testmodelrelated'),
        ),
        migrations.AlterField(
            model_name='testmodel',
            name='related_through',
            field=models.ManyToManyField(blank=True, related_name='+', through='test_app.TestModelThrough', to='test_app.testmodelrelated'),
        ),
    ]
